import sys
from typing import List, Dict, Union

import mako
import mako.exceptions
import mako.lookup
import mako.template
import os

from pydmt.api.builder import Builder
from pydmt.utils.filesystem import makedirs_for_file
from pydmt.utils.digest import sha1_files


class Mako(Builder):
    def get_sources(self) -> List[str]:
        return [self.source]

    def __init__(self,
                 source: str,
                 target: str,
                 data: Union[Dict[str, object], None],
                 dep_files: List[str],
                 ):
        super().__init__()
        self.source = source  # type: str
        self.target = target  # type: str
        self.dep_files = dep_files  # type: List[str]
        if data is None:
            self.data = dict()
        else:
            self.data = data

    def get_signature(self) -> str:
        # FIXME: this should be the sha1 of the source + all the definition files
        # and even a better correction:
        # sha1 of the source file + the sha1 of all the variables references from within
        # the source file.
        return sha1_files(self.dep_files+[self.source])

    def build(self):
        try:
            template = mako.template.Template(filename=self.source)
            makedirs_for_file(self.target)
            with open(self.target, 'w') as file_handle:
                file_handle.write(template.render(**self.data))
        except Exception as e:
            if os.path.isfile(self.target):
                os.unlink(self.target)
            raise e

    def get_targets(self) -> List[str]:
        return [self.target]


def print_full_exception():
    print('printing full exception')
    traceback = mako.exceptions.RichTraceback()
    for (filename, line_number, function_name, line) in traceback.traceback:
        print('File {0}, line {1}, in {2}'.format(filename, line_number, function_name))
        print(line)
    print('{0}: {1}'.format(str(traceback.error.__class__.__name__), traceback.error))


def print_exception(e, input_file):
    found = False
    traceback = mako.exceptions.RichTraceback()
    for (filename, line_number, function_name, line) in traceback.traceback:
        if filename == input_file:
            print('{0}: error {1} in {2}, line {3}'.format(sys.argv[0], str(e), filename, line_number, function_name))
            print('{0}'.format(line))
            found = True
    if not found:
        for (filename, line_number, function_name, line) in traceback.traceback:
            print('File {0}, line {1}, in {2}'.format(filename, line_number, function_name))
            print(line)
        print('{0}: {1}'.format(str(traceback.error.__class__.__name__), traceback.error))



import os
import pylogconf.core
import sys
import pathlib

from pydmt.builders.sphinx import Sphinx
from pydmt.core.pydmt import PyDMT
from pydmt.features.templating import Templating


def main():
    # parameters
    debug = False
    sphinx = True
    add_import_of_cwd = True
    add_import_of_home = True
    add_import_of_shared = True
    if add_import_of_shared:
        folder = "/etc/pydmt"
        if folder not in sys.path:
            sys.path.insert(0, folder)
    if add_import_of_home:
        folder = os.path.join(str(pathlib.Path.home()), ".config/pydmt")
        if folder not in sys.path:
            sys.path.insert(0, folder)
    if add_import_of_cwd:
        if "" not in sys.path:
            sys.path.insert(0, "")

    pylogconf.core.setup()
    if debug:
        print("system path is [{}]".format(sys.path))
    p = PyDMT()

    # add templating support
    t = Templating()
    t.setup(p)

    # add sphinx builder support
    if sphinx and os.path.isdir("sphinx"):
        import config.project
        p.add_builder(Sphinx(package_name=config.project.project_name))
    p.build_all()


if __name__ == '__main__':
    main()



def to_php(x):
    if type(x) == str:
        return '\'{0}\''.format(x)
    if type(x) == bool:
        if x:
            return 'TRUE'
        else:
            return 'FALSE'
    if x is None:
        return 'null'
    raise ValueError('dont know how to translate type', type(x), x)



import logging
from typing import List, Dict

import os

from pydmt.api.builder import Builder
from pydmt.core.cache import Cache
from pydmt.utils.filesystem import copy_mkdir
from pydmt.utils.digest import sha1_file


class BuildProcessStats:
    def __init__(self):
        self.builder_fail = []
        self.builder_ok = []
        self.copy_missing = []
        self.copy_sha1 = []
        self.nop = []

    def add_copy_missing(self, filename: str, object_name: str):
        self.copy_missing.append((filename, object_name))

    def get_copy_missing(self):
        return len(self.copy_missing)

    def add_copy_sha1(self, filename: str, object_name: str):
        self.copy_sha1.append((filename, object_name))

    def add_nop(self, filename: str, object_name: str):
        self.nop.append((filename, object_name))

    def get_nop(self):
        return len(self.nop)

    def add_builder_ok(self, builder: Builder):
        self.builder_ok.append(builder)

    def add_builder_fail(self, builder: Builder, e: Exception):
        self.builder_fail.append((builder, e))

    def get_builder_ok(self):
        return len(self.builder_ok)

    def get_builder_fail(self):
        return len(self.builder_fail)


class PyDMT:
    def __init__(self):
        self.builders = []  # type: List[Builder]
        self.target_to_builder = {}  # type: Dict[str, Builder]
        self.cache = Cache()

    def build_by_builder(self, builder: Builder, stats: BuildProcessStats):
        """ run one builder, return statistics about the run """
        logger = logging.getLogger(__name__)
        target_signature = builder.get_signature()
        assert target_signature is not None, "builder signature is None"
        if self.cache.list_sig_ok(target_signature):
            logger.info("verifying [{}]".format(builder.get_name()))
            file_bad = 0
            file_correct = 0
            file_missing = 0
            file_total = 0
            list_filename = self.cache.get_list_filename(target_signature)
            for object_name, signature in Cache.iterate_objects(list_filename):
                filename = self.cache.get_object_filename(signature)
                if os.path.isfile(object_name):
                    object_name_signature = sha1_file(object_name)
                    if object_name_signature != signature:
                        # logger.info("file [{}] is incorrect. Getting from cache.".format(object_name))
                        copy_mkdir(filename, object_name)
                        stats.add_copy_sha1(filename, object_name)
                        file_bad += 1
                    else:
                        # logger.info("file [{}] is up to date".format(object_name))
                        stats.add_nop(filename, object_name)
                        file_correct += 1
                else:
                    # logger.info("file [{}] is missing. Getting from cache.".format(object_name))
                    copy_mkdir(filename, object_name)
                    stats.add_copy_missing(filename, object_name)
                    file_missing += 1
                file_total += 1
            if file_bad > 0 or file_missing > 0:
                logger.info("Retrieved {} files from cache (bad/correct/missing = {}/{}/{}".format(
                    file_total, file_bad, file_correct, file_missing))
            else:
                logger.info("ok [{}]".format(builder.get_name()))
        else:
            # this is one of the rare cases in which really want to catch all exceptions.
            # noinspection PyBroadException
            try:
                logger.info("running [{}]".format(builder.get_name()))
                builder.build()
                logger.info("ok [{}]".format(builder.get_name()))
                stats.add_builder_ok(builder)
                # first lets build a list of what was constructed
                targets = builder.get_targets()
                targets.extend(builder.get_targets_post_build())
                content = ""
                for target in targets:
                    signature = sha1_file(target)
                    content += target + " " + signature + "\n"
                    self.cache.save_object_by_signature(signature, target)
                self.cache.save_list_by_signature(target_signature, content)
            except Exception as e:
                logger.info("failed [{}]".format(builder.get_name()))
                logger.info("exception [{}]".format(e))
                stats.add_builder_fail(builder, e)

    def build_by_target(self, target: str, stats: BuildProcessStats) -> None:
        b = self.target_to_builder[target]
        self.build_by_builder(b, stats)

    def build_by_targets(self, targets: List[str], stats: BuildProcessStats) -> None:
        for target in targets:
            self.build_by_target(target, stats)

    def build_all(self) -> BuildProcessStats:
        """
        Build all the targets, very high level method
        :return:
        """
        stats = BuildProcessStats()
        for builder in self.builders:
            self.build_by_builder(
                builder=builder,
                stats=stats,
            )
        return stats

    def add_builder(self, b: Builder) -> None:
        self.builders.append(b)
        targets = b.get_targets()
        if targets:
            for target in targets:
                self.target_to_builder[target] = b



import setuptools

"""
The documentation can be found at:
http://setuptools.readthedocs.io/en/latest/setuptools.html
"""
setuptools.setup(
    # the first three fields are a must according to the documentation
    name='pydmt',
    version='0.0.54',
    packages=[
        'pydmt',
        'pydmt.api',
        'pydmt.builders',
        'pydmt.core',
        'pydmt.features',
        'pydmt.scripts',
        'pydmt.utils',
    ],
    # from here all is optional
    description='python dependency management tool',
    long_description='python dependency management tool',
    author='Mark Veltzer',
    author_email='mark.veltzer@gmail.com',
    maintainer='Mark Veltzer',
    maintainer_email='mark.veltzer@gmail.com',
    keywords=[
        'pydmt',
        'cons',
        'scons',
        'software construction',
        'make',
        'maven',
        'mvn',
    ],
    url='https://veltzer.github.io/pydmt',
    download_url='https://github.com/veltzer/pydmt',
    license='MIT',
    platforms=[
        'python3',
    ],
    install_requires=[
        'pyfakeuse',
        'pylogconf',
        'mako',
        'Sphinx',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities',
    ],
    data_files=[
    ],
    entry_points={'console_scripts': [
        'pydmt_build=pydmt.scripts.build:main',
    ]},
    python_requires='>=3',
)
